#!/bin/bash
# This script is used to stop the celery workers
# Automatically generated by acoupi, DO NOT EDIT THIS FILE
# Date: {{ now }}

{{ celery_bin }} \
	-A {{ app_name }} \
	multi \
	stopwait \
	{% for worker in config.workers %}{{ worker.name }} {% endfor %} \
	--pidfile={{ run_dir }}/%n.pid \
	--logfile={{ log_dir }}/%n%I.log \
	--loglevel="{{ log_level }}"
