[Unit]
Description=Acoupi Service
After=network.target

[Service]
Type=forking
EnvironmentFile={{ environment_file }}
WorkingDirectory={{ working_directory }}
ExecStart={{ start_script }}
ExecStop={{ stop_script }}
ExecReload={{ restart_script }}
Restart=always
RestartSec=3

[Install]
WantedBy=default.target
