[Unit]
Description=Acoupi Scheduler Service
After=network.target

[Service]
Type=simple
EnvironmentFile={{ environment_file }}
WorkingDirectory={{ working_directory }}
ExecStart={{ beat_script }}
Restart=always
RestartSec=3

[Install]
WantedBy=default.target
